/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.pinot.core.segment.name;

import com.google.common.base.Preconditions;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.TimeZone;
import java.util.concurrent.TimeUnit;
import javax.annotation.Nullable;
import org.apache.pinot.common.data.TimeGranularitySpec.TimeFormat;


/**
 * Segment name generator that normalizes the date to human readable format.
 */
public class NormalizedDateSegmentNameGenerator implements SegmentNameGenerator {
  private final String _segmentNamePrefix;
  private final boolean _appendPushType;
  private final boolean _excludeSequenceId;

  // For APPEND tables
  private final SimpleDateFormat _outputSDF;
  // For EPOCH time format
  private final TimeUnit _inputTimeUnit;
  // For SIMPLE_DATE_FORMAT time format
  private final SimpleDateFormat _inputSDF;

  public NormalizedDateSegmentNameGenerator(String tableName, @Nullable String segmentNamePrefix,
      @Nullable String excludeSequenceId, @Nullable String pushType, @Nullable String pushFrequency,
      @Nullable String timeType, @Nullable String timeFormat) {
    _segmentNamePrefix = segmentNamePrefix != null ? segmentNamePrefix.trim() : tableName;
    _appendPushType = "APPEND".equalsIgnoreCase(pushType);
    _excludeSequenceId = Boolean.parseBoolean(excludeSequenceId);

    // Include time info for APPEND push type
    if (_appendPushType) {
      // For HOURLY push frequency, include hours into output format
      if ("HOURLY".equalsIgnoreCase(pushFrequency)) {
        _outputSDF = new SimpleDateFormat("yyyy-MM-dd-HH");
      } else {
        _outputSDF = new SimpleDateFormat("yyyy-MM-dd");
      }
      _outputSDF.setTimeZone(TimeZone.getTimeZone("UTC"));

      // Parse input time format: 'EPOCH' or 'SIMPLE_DATE_FORMAT:<pattern>'
      if (Preconditions.checkNotNull(timeFormat).equals(TimeFormat.EPOCH.toString())) {
        _inputTimeUnit = TimeUnit.valueOf(timeType);
        _inputSDF = null;
      } else {
        Preconditions.checkArgument(timeFormat.startsWith(TimeFormat.SIMPLE_DATE_FORMAT.toString()),
            "Invalid time format: %s, must be one of '%s' or '%s:<pattern>'", timeFormat, TimeFormat.EPOCH,
            TimeFormat.SIMPLE_DATE_FORMAT);
        _inputTimeUnit = null;
        _inputSDF = new SimpleDateFormat(timeFormat.substring(timeFormat.indexOf(':') + 1));
        _inputSDF.setTimeZone(TimeZone.getTimeZone("UTC"));
      }
    } else {
      _outputSDF = null;
      _inputTimeUnit = null;
      _inputSDF = null;
    }
  }

  @Override
  public String generateSegmentName(int sequenceId, @Nullable Object minTimeValue, @Nullable Object maxTimeValue) {
    Integer sequenceIdInSegmentName = !_excludeSequenceId && sequenceId >= 0 ? sequenceId : null;
    if (!_appendPushType) {
      return JOINER.join(_segmentNamePrefix, sequenceIdInSegmentName);
    } else {
      return JOINER.join(_segmentNamePrefix, getNormalizedDate(Preconditions.checkNotNull(minTimeValue)),
          getNormalizedDate(Preconditions.checkNotNull(maxTimeValue)), sequenceIdInSegmentName);
    }
  }

  /**
   * Converts the time value into human readable date format.
   *
   * @param timeValue Time value
   * @return Normalized date string
   */
  private String getNormalizedDate(Object timeValue) {
    if (_inputTimeUnit != null) {
      return _outputSDF.format(new Date(_inputTimeUnit.toMillis(Long.parseLong(timeValue.toString()))));
    } else {
      try {
        return _outputSDF.format(_inputSDF.parse(timeValue.toString()));
      } catch (ParseException e) {
        throw new RuntimeException(String
            .format("Caught exception while parsing simple date format: %s with value: %s", _inputSDF.toPattern(),
                timeValue), e);
      }
    }
  }

  @Override
  public String toString() {
    StringBuilder stringBuilder =
        new StringBuilder("NormalizedDateSegmentNameGenerator: segmentNamePrefix=").append(_segmentNamePrefix)
            .append(", appendPushType=").append(_appendPushType);
    if (_excludeSequenceId) {
      stringBuilder.append(", excludeSequenceId=true");
    }
    if (_outputSDF != null) {
      stringBuilder.append(", outputSDF=").append(_outputSDF.toPattern());
    }
    if (_inputTimeUnit != null) {
      stringBuilder.append(", inputTimeUnit=").append(_inputTimeUnit);
    }
    if (_inputSDF != null) {
      stringBuilder.append(", inputSDF=").append(_inputSDF.toPattern());
    }
    return stringBuilder.toString();
  }
}
